//                              -*- Mode: C++ -*- 
// ppm2avi.cc -- create an uncompressed movie out of a slew of PPMs. Adapted
// from avimake. Main use to post-process ppm files from pov output and
// avirecompress....
// 
// 
// $Id: avimake.cpp,v 1.7 2001/10/05 11:02:48 kabi Exp $
// 
// Copyright (C) 2001  Tom Pavel <pavel@alum.mit.edu>
// Copyright (C) 2001  Nicolas Brodu <nicolas.brodu@free.fr>
// 
// Creator         : Tom Pavel  <tom>     Sun Feb  4 01:23:17 2001
// Modifier        : Tom Pavel  <tom>     Sun May 13 00:35:21 2001
// Update Count    : 45
// Status          : Unknown, Use with caution!
// 


#include "avifile.h"
#include "fourcc.h"
#include "except.h"
#include "aviplay.h"
#include "version.h"

#include <unistd.h>		// for getopt()
#include <string.h>		// for memset()
#include <stdlib.h>		// atoi()
#include <errno.h>
#include <stdio.h>
#include <sys/types.h>
#include <assert.h>
#include <ctype.h>

#include <iostream>

using namespace std;


// Public domain Berkeley's code from ppmtoyuv.c in mpeg_encode package
// Adapted to store data in a contiguous buffer

typedef	unsigned char uint8;
typedef char int8;
typedef int boolean;
#define TRUE 1
#define FALSE 0


static int	    height, width;
static uint8   *ppm_data = 0;
static uint8 **orig_y, **orig_cr, **orig_cb;


/*=====================*
 * INTERNAL PROCEDURES *
 *=====================*/

static int	ReadNextInteger(FILE *fpointer)
{
    char    input[256];
    int	    index;

    /* skip whitespace */
    while ( fgets(input, 2, fpointer) != NULL ) 
    {
	if ( isspace(input[0]) )
	    continue;

	if ( input[0]=='#' )	
	/* read the comment lines from # to \n */
	   {
	      while ( fgets(input, 2, fpointer) != NULL )
	       {
		if(input[0]=='\n') 
			break;
	        }
	     continue;
	    }


	/* read rest of integer */
	index = 1;
	while ( fgets(&input[index], 2, fpointer) != NULL )
	{
	    if ( isspace(input[index]) )
		break;
	    index++;
	}
	input[index] = '\0';

	return atoi(input);
    }

    return -1;	    /* end of file reached */
}

static bool alreadyRead = false;

static boolean	ReadPPM(FILE *fpointer)
{
    char    input[256];
    uint8   junk[4096];
    register int y;
    int  maxVal;

    if ( fread(input, sizeof(char), 2, fpointer) != 2 )
	return FALSE;

    if ( strncmp(input, "P6", 2) != 0 )	    /* magic number */
	return FALSE;

    width = ReadNextInteger(fpointer);
    if ( width == -1 )
        return FALSE;

    height = ReadNextInteger(fpointer);
    if ( height == -1 )
	return FALSE;

    maxVal = ReadNextInteger(fpointer);
    if ( maxVal == -1 )
	return FALSE;

    if ( maxVal != 255 )
    {
	fprintf(stdout, "MAXVAL != 255!!!  Exiting!\n");
	fprintf(stdout, "MAXVAL war %d\n",maxVal);
	exit(1);
    }

	if (!ppm_data) 
		ppm_data = (uint8 *) malloc(sizeof(uint8) * width * height * 3);


	uint8* dat = ppm_data;
	int x;
    for ( y = 0; y < height; y++ )
	    for ( x = 0; x < width; x++ ) {
			fread(dat, sizeof(char), 3, fpointer);
			uint8 tmp = dat[0]; // swap B/R values
			dat[0] = dat[2];
			dat[2] = tmp;
			dat += 3;
		}

	alreadyRead = true;
    return TRUE;
}





void
fillInfo (const char* fname, BITMAPINFOHEADER* bi)
{
    // open and read fname as PPM file

    FILE* infile = fopen(fname, "rb");
    if (infile  == NULL) {
	cerr << "can't open " <<  fname
	     << ": " << strerror(errno) << endl;
	exit(1);
    }

	if (!ReadPPM(infile)) {
		cerr << "could not read PPM file " << fname << endl;
		exit(1);
	}


    memset(bi, 0, sizeof(*bi));
    bi->biSize = sizeof(*bi);
    bi->biWidth = width;
    bi->biHeight = height;
    bi->biSizeImage = bi->biWidth * bi->biHeight * 3;
    bi->biPlanes = 1;
    bi->biBitCount = 24;

    fclose(infile);
}


void
addFrame (const char* fname, IAviVideoWriteStream* outVidStr)
{

	if (!alreadyRead) {
	    FILE* infile = fopen(fname, "rb");
    	if (infile  == NULL) {
		cerr << "can't open " <<  fname
		     << ": " << strerror(errno) << endl;
		exit(1);
    	}

		if (!ReadPPM(infile)) {
			cerr << "could not read PPM file " << fname << endl;
			exit(1);
		}
	}

    // create frame from 24-bit image array.
    CImage frame(ppm_data, width, height);

    outVidStr->AddFrame(&frame);
	
	alreadyRead = false;
}



void
Usage (const char* progname)
{
    cerr << "usage: " << progname
	 << " [-h] [-o outfile] [-r fr] [-e codec]" << endl
	 << "With: fr: frame rate. Default is 24 fps." << endl
	 << "      codec: 4-character codec identifier. Default is DIVX." << endl;
    exit(1);
}


int
main (int argc, char* argv[])
try
{
    int debug = 0;
    const char* outFn = "out.avi";
    const char* encoder = "DIVX";
    int fr = 24;			// default to 24 fps

    int ch;
    while ((ch = getopt(argc, argv, "dho:r:e:")) != EOF) {
        switch ((char)ch) {
        case 'd':
            ++debug;
	    break;

	case 'o':
	    outFn = optarg;
	    break;

	case 'e':
	    encoder = optarg;
 	    if (strlen(encoder)!=4)
		Usage (argv[0]);
	    break;

	case 'r':
	    fr = atoi (optarg);
	    if (fr == 0)
		Usage (argv[0]);
	    break;
	    
	case 'h':
        case '?':
        default:
	    Usage (argv[0]);
	}
    }

    argc -= optind;
    argv += optind;
    if (argc < 1) Usage (argv[0]);


    // Standard AVIlib sanity check:
    if ( GetAvifileVersion() != AVIFILE_VERSION) {
        cout << "This binary was compiled for Avifile ver. "
	     << AVIFILE_VERSION
	     << ", but the library is ver. "
	     << GetAvifileVersion()
	     << ". Aborting."
	     << endl;
        return 0;
    }   

    // Do the real work
    IAviWriteFile* outFile = CreateSegmentedFile(outFn);
    BITMAPINFOHEADER bi;
	
    //FOURCC codec = fccDIV3;
    //fourcc_t codec = fccMP42;
    //FOURCC codec = fccIV32;
    //FOURCC codec = fccCVID;

	// fourcc_t codec  = mmioFOURCC('R', 'G', 'B', ' ');
    
	fourcc_t codec  = mmioFOURCC(encoder[0], encoder[1], encoder[2], encoder[3]);
	
//	fourcc_t codec  = RIFFINFO_DIVX;

    // Use the first file to pick the image sizes:
    fillInfo(*argv, &bi);

    IAviVideoWriteStream* vidStr =
	outFile->AddVideoStream(codec, &bi, 1000000/fr);

    vidStr->Start();
    while (argc > 0) {
		cout << "processing " << *argv << "...";
		addFrame (*argv, vidStr);
		cout << " done." << endl;
        --argc;
        ++argv;	
    }
    vidStr->Stop();

    // Close the outFile and write out the header, etc.
    delete outFile;
	free(ppm_data);
}

catch(FatalError& error) {
    error.Print();
}   
